//Copyright (c) 2011, Paul Lindt, Irina Lindt
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    //Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    //Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or
//other materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
//DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
//USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE 



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 

#include "linkedlist.h"

#define CLIENT_MODE 0
#define SERVER_MODE 1
#define MAX_CLIENTS 5
#define MAX_INPUT	50
#define MAX_INPUT_LENGTH 50
#define PORT 9020

#define FALSE 0
#define TRUE 1



struct on_connection
{
	int on_socket;
	struct thread_list *server_threads;
	struct socket_list *sockets;
	int client_number;
	
	char username [MAX_USERNAME_SIZE + 1];
};

//structur wird nur gebraucht um beide variablen an pthread_create als
//parameter zu uebergeben
struct tmp_socket_struct
{
	int socket;
	struct on_connection * connection;
};

void on_inform_clients(struct on_connection * my_connection, char * input)
{
		#ifdef DEBUG
		printf("%s\n", "informiere clienten");
		#endif

		struct socket_list * tmp = my_connection->sockets;
		while(tmp != NULL)
		{
			int socket = tmp->socket;
			#ifdef DEBUG
				printf("Informiere socket %d\n", socket);
			#endif
			send(socket, input, MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2,0);
			tmp = tmp->next;
		}
		
}



//reagiert auf von einem client kommende Informationen
void client_handler(void *adsf)
{
	struct on_connection * my_connection = ((struct tmp_socket_struct *)adsf)->connection;
	int thread_socket = ((struct tmp_socket_struct *)adsf)->socket;
	char input[MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2];
	#ifdef DEBUG
		printf("socket: %d\n", thread_socket);
	#endif
	while(TRUE)
	{
		if(recv(thread_socket, input, MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2 , 0) == 0)
		{
			struct socket_list * tmp = my_connection->sockets;
			
			//~ #ifdef DEBUG
			printf("%s", "Client beendet");
			//~ #endif
			if(my_connection->sockets->socket == thread_socket)
			{
				my_connection->sockets = my_connection->sockets->next;
				my_connection->sockets->next;
				free(tmp);
			}else{
				while(tmp->next != NULL)
				{
					if(tmp->next->socket == thread_socket)
					{
						struct socket_list *tmp2 = tmp->next;
						tmp->next = tmp->next->next;
						free(tmp2);
					}
				}
			}
			pthread_exit(0);
			
		}
		on_inform_clients(my_connection, input);
		#ifdef DEBUG
		printf("input:%s, length: %d\n", input, (int)strlen(input));
		#endif
	}
	
	
}

int on_connect(struct on_connection *my_connection, int mode)
{
	my_connection->on_socket = socket(AF_INET, SOCK_STREAM, 0);
	struct sockaddr_in server;
	struct hostent *remote_server = gethostbyname("localhost");
	memcpy(&server.sin_addr, remote_server->h_addr_list[0], remote_server->h_length);
	server.sin_family = AF_INET;
	server.sin_port = htons(PORT);
	if(mode == CLIENT_MODE)
	{
		connect(my_connection->on_socket, (struct sockaddr *)&server, sizeof(server));
	}
	else
	{
		bind(my_connection->on_socket, (struct sockaddr *)&server, sizeof(server));
		listen(my_connection->on_socket, 5);
		#ifdef DEBUG	
			printf("%s", "erstellen der Listen");
		#endif
		
		my_connection->sockets=NULL;
		my_connection->server_threads = NULL;
	}
	
	
	
	return 0;
}


	

//schreibt eine neue nachricht ins schared memory
//informiert den server darueber
void on_send(struct on_connection * my_connection, char *input)
{
	
	
	char * message = malloc(strlen(my_connection->username) + 2 + strlen(input));
	strcpy(message, my_connection->username);
	strcpy(&message[strlen(my_connection->username)], ": ");
	strcpy(&message[strlen(my_connection->username) + 1],input);
	
	#ifdef DEBUG
		printf("sende Nachricht:%s\n", message);
	#endif
	
	
	send(my_connection->on_socket, message, MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2 , 0);
}
void on_recv(struct on_connection * my_connection)
{
	char input[MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2];
	if(recv(my_connection->on_socket, input, MAX_INPUT_LENGTH + MAX_USERNAME_SIZE + 2 , 0) == 0)
			error("Keine Verbingung zum Server");
	printf("%s\n", input);
	#ifdef DEBUG
		printf("length: %d\n", (int)strlen(input));
	#endif
	
}	


void on_disconnect(struct on_connection * my_connection,int mode)
{
	close(my_connection->on_socket);
}


void ond_server_interface(void *my_connection_void)
{
	#ifdef DEBUG
		printf("%s\n", "im server thread");
	#endif
	
	struct on_connection * my_connection = (struct on_connection *)my_connection_void;
	int on_tmp_socket;
	
	while(TRUE)
	{
		on_tmp_socket = accept(my_connection->on_socket, NULL, 0);
		//~ socketAddElement(my_connection->sockets,on_tmp_socket);
		struct socket_list * tmp_link;
		tmp_link = malloc(sizeof(struct socket_list));
		tmp_link->next = my_connection->sockets;
		tmp_link->socket = on_tmp_socket;
		my_connection->sockets = tmp_link;
		
		struct tmp_socket_struct adsf;
		adsf.socket = on_tmp_socket;
		adsf.connection = my_connection;
		pthread_t * tmp_server_thread = malloc(sizeof(pthread_t));
		if((pthread_create(tmp_server_thread, NULL,(void *)client_handler, (void *)&adsf) != 0))
		{
			error("konnte den client_handler_thread nicht erstellen");
		}
		my_connection->client_number = my_connection->client_number + 1;
		threadAddElement(my_connection->server_threads, tmp_server_thread);
	}
		

	pthread_exit(NULL);
}
