//Copyright (c) 2011, Paul Lindt, Irina Lindt
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    //Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    //Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or
//other materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
//DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
//USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE 




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>


#include <sys/mman.h>
#include <sys/stat.h>        /* For mode constants */
#include <fcntl.h>           /* For O_* constants */

#include <unistd.h>


#define SERVER_MODE (O_CREAT | O_RDWR)
#define CLIENT_MODE	O_RDWR

struct on_sync_vars
{
	pthread_mutex_t server_mutex, client_mutex, log_mutex;
	pthread_cond_t server_blocking_cv, clients_blocking_cv;
};

struct on_connection
	{
		int shm_log_descriptor;
		int shm_sync_vars_descriptor;
		struct on_sync_vars * sync_vars;
		
		char username [MAX_USERNAME_SIZE + 1] ;
	};

	





void write_cv(struct on_connection *my_connection)
{
	pthread_mutexattr_t mutex_attr;
	pthread_mutexattr_init(&mutex_attr);
	pthread_mutexattr_setpshared(&mutex_attr, PTHREAD_PROCESS_SHARED);
	
	pthread_condattr_t condvar_attr;
	pthread_condattr_setpshared(&condvar_attr, PTHREAD_PROCESS_SHARED);
	
	
	pthread_mutex_init(&(my_connection->sync_vars->server_mutex), &mutex_attr);
	pthread_mutex_init(&(my_connection->sync_vars->client_mutex), &mutex_attr);
	pthread_cond_init(&(my_connection->sync_vars->server_blocking_cv), &condvar_attr);
	pthread_cond_init(&(my_connection->sync_vars->clients_blocking_cv), &condvar_attr);

	
	pthread_mutex_init(&(my_connection->sync_vars->log_mutex),&mutex_attr);
	pthread_mutexattr_destroy(&mutex_attr);
}
	

//oeffnet die shared memory
void on_open_shm(struct on_connection * my_connection, int mode)
{
	if((my_connection->shm_sync_vars_descriptor = shm_open("./on_sync_vars", mode, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) == 0)
		error("konnte die shared memory sync_vars datei nicht oeffnen");
	if((my_connection->shm_log_descriptor = shm_open("./on_log", mode, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) == 0)
		error("konnte die shared memory log datei nicht oeffnen");	
		
	fchmod(my_connection->shm_sync_vars_descriptor, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	fchmod(my_connection->shm_log_descriptor, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	//aus irgendeinemgrund werden die rechte beim erstellen der Messagequeue und der Sharedmemory nicht richtig gesetzt
	//deshalb werden sie hier nochmal gesetzt.
	
	ftruncate(my_connection->shm_sync_vars_descriptor, sizeof(struct on_sync_vars));
		
			
	if((my_connection->sync_vars = mmap(NULL, 	//der Kernel darf die Adresse bestimmen
								sizeof(struct on_sync_vars),		//groesse
								PROT_READ | PROT_WRITE,
								MAP_SHARED,
								my_connection->shm_sync_vars_descriptor,
								0)) == MAP_FAILED)
								{
			error("konnte datei nicht im speicher mappen");
		}
}



int on_connect(struct on_connection* my_connection, int mode)
{
	on_open_shm(my_connection, mode);
	if(mode == SERVER_MODE)
	{
		write_cv(my_connection);
	}
	return 0;
}


	
	

//schreibt eine neue nachricht ins schared memory
//informiert den server darueber
void on_send(struct on_connection * my_connection, char *input)
{
	#ifdef DEBUG
		printf("sende input:%s\n", input);
	#endif
	
	//ins log schreiben
	pthread_mutex_lock(&my_connection->sync_vars->log_mutex);
	write(my_connection->shm_log_descriptor, my_connection->username, strlen(my_connection->username));
	write(my_connection->shm_log_descriptor, ": ", 2);
	write(my_connection->shm_log_descriptor, input, strlen(input));
	pthread_mutex_unlock(&my_connection->sync_vars->log_mutex);
	
	
	//server informieren
	pthread_mutex_lock(&my_connection->sync_vars->server_mutex);
	pthread_cond_signal(&my_connection->sync_vars->server_blocking_cv);
	pthread_mutex_unlock(&my_connection->sync_vars->server_mutex);
}

void ond_send(struct on_connection * my_connection, char *input)
{
	#ifdef DEBUG
		printf("sende input:%s\n", input);
	#endif
	
	//ins log schreiben
	pthread_mutex_lock(&my_connection->sync_vars->log_mutex);
	write(my_connection->shm_log_descriptor, input, strlen(input));
	pthread_mutex_unlock(&my_connection->sync_vars->log_mutex);
	
	
	//server informieren
	pthread_mutex_lock(&my_connection->sync_vars->client_mutex);
	pthread_cond_broadcast(&my_connection->sync_vars->clients_blocking_cv);
	pthread_mutex_unlock(&my_connection->sync_vars->client_mutex);
}




void on_recv(struct on_connection * my_connection)
{
	lseek(my_connection->shm_log_descriptor, 0l, SEEK_END);
	
	//warten auf eingabe
	pthread_mutex_lock(&my_connection->sync_vars->client_mutex);
	pthread_cond_wait(&my_connection->sync_vars->clients_blocking_cv, 
	&my_connection->sync_vars->client_mutex);
	
	//log anzeigen
	pthread_mutex_lock(&my_connection->sync_vars->log_mutex);
	struct stat filestat;
	fstat(my_connection->shm_log_descriptor, &filestat);
	char *input;
	input = (char *)malloc(filestat.st_size + 1);
	lseek(my_connection->shm_log_descriptor, 0l, SEEK_SET);
	read(my_connection->shm_log_descriptor, input, filestat.st_size );
	input[filestat.st_size] = 0;
	
	pthread_mutex_unlock(&my_connection->sync_vars->log_mutex);
	printf("%s\n", "-------------------------------Unterhaltung---------------------------------");
	printf("%s\n", input);
	#ifdef DEBUG
		printf("log groesse: %d\n",(int)filestat.st_size);
	#endif
	free(input);
	pthread_mutex_unlock(&my_connection->sync_vars->client_mutex);
}	


void on_inform_clients(struct on_connection * my_connection)
{
	pthread_mutex_lock(&my_connection->sync_vars->server_mutex);
	pthread_cond_wait(&my_connection->sync_vars->server_blocking_cv, &my_connection->sync_vars->server_mutex);
	
	#ifdef DEBUG	
	printf("%s\n", "informiere clienten");
	#endif
	
	pthread_mutex_lock(&my_connection->sync_vars->client_mutex);
	pthread_cond_broadcast(&my_connection->sync_vars->clients_blocking_cv);
	//daraufhin broadcast fuer die andere
	pthread_mutex_unlock(&my_connection->sync_vars->client_mutex);
	pthread_mutex_unlock(&my_connection->sync_vars->server_mutex);	
	
}

void on_disconnect(struct on_connection * my_connection,int mode)
{
	if(mode == SERVER_MODE)
	{
		printf("%s\n", "server wird beendet");
		ond_send(my_connection, "server wird beendet\n\n\n\n\n");
		shm_unlink("/dev/shm/on_log");
		shm_unlink("/dev/shm/on_sync_vars");
		remove("/dev/shm/on_log");
		remove("/dev/shm/on_sync_vars");
	}
	else
	{
		;
	}
}


void ond_server_interface(void *my_connection)
{
	#ifdef DEBUG
		printf("%s\n", "im server thread");
	#endif
	
	int done = FALSE;
	while(!done)
	{
		//auf eine condvar warten
		
		
		on_inform_clients((struct on_connection *)my_connection);
		
		
	}


	pthread_exit(NULL);
}
